#include "shadermgr.h"
#include "UTIL_openGL.h"
#include "Log.h"
#define _ERROR_REPORT

int CheckGLError(char *file, int line)
{
	GLenum glErr;
	int    retCode = 0;
	char sz[256];
	glErr = glGetError();
	while (glErr != GL_NO_ERROR) {
		sprintf(sz,"GL Error #%i (%s) in File %s at line: %i\n",glErr,gluErrorString(glErr),file,line);
        gLog.OutPut(sz);
		retCode = 1;
		glErr = glGetError();
	}
	return retCode;
}


ShaderMgr::ShaderMgr(void)
{
	curIndex =0;
	mCurProgram = 0;
}

ShaderMgr::~ShaderMgr(void)
{
}

unsigned int ShaderMgr::CreateShader(int type,  char* data)
{
	if(!data)
		return 0;

	Shader	s;
	s.handle = glCreateShaderObjectARB(type);
	int shaderlength=(int)strlen(data);
	glShaderSourceARB(s.handle,1,(const GLcharARB**)&data,&shaderlength);
	glCompileShaderARB(s.handle);
	int status=0;
	glGetObjectParameterivARB(s.handle, GL_OBJECT_COMPILE_STATUS_ARB, &status);
	UTIL_GL::GetGLError();
	if(!status)
	{
		gLog.OutPut("Failed To Compile Shader!\n");
#ifdef _ERROR_REPORT
		int len =0,loglen =0;
	    char *log;

		glGetObjectParameterivARB(s.handle,GL_OBJECT_INFO_LOG_LENGTH_ARB,&len);
	    if (len > 0)
	    {
			log = (char *)malloc(len);
			glGetInfoLogARB(s.handle, len, &loglen, log);
			gLog.OutPut(log);
			gLog.OutPut("\n");
			free(log);
	    }
#endif
		return 0;
	}
	s.index = ++curIndex;
	mShaders.push_back(s);
	return s.index;
}
unsigned int ShaderMgr::CreateProgram()
{
	Program p;
	p.handle = glCreateProgramObjectARB();
	p.index = ++curIndex;
	mPrograms.push_back(p);
	mCurProgram = &mPrograms.back();
	return p.index;
}
Shader* ShaderMgr::FindShader(unsigned int idx)
{
	vector<Shader>::iterator s;
	for(s=mShaders.begin();s<mShaders.end();s++)
	{
		if((*s).index == idx)
			return &(*s);
	}
	return 0;
}
Program* ShaderMgr::FindProgram(unsigned int idx)
{
	vector<Program>::iterator p;
	for(p=mPrograms.begin();p<mPrograms.end();p++)
	{
		if((*p).index == idx)
			return &(*p);
	}
	return 0;
}
void ShaderMgr::AttachShader(unsigned int shader)
{
	Shader* s=NULL;
	if(s=FindShader(shader))
	{
		glAttachObjectARB(mCurProgram->handle,s->handle);
	}
}
void ShaderMgr::DetachShader(unsigned int shader)
{
	Shader* s=NULL;
	if(s=FindShader(shader))
	{
		glDetachObjectARB(mCurProgram->handle,s->handle);
	}
}
void ShaderMgr::DeleteProgram(unsigned int prg)
{
	Program* p=NULL;
	if(p=FindProgram(prg))
	{
		vector<Program>::iterator s;
		for(s=mPrograms.begin();s<mPrograms.end();s++)
		{
			if((*s).index == p->index)
			{
				glDetachObjectARB(p->handle, (*s).handle);
				glDeleteShader((*s).handle);
				mPrograms.erase(s);
				return;
			}
		}
		glDeleteObjectARB(p->handle);
	}

}
bool ShaderMgr::LinkProgram()
{
	glLinkProgramARB(mCurProgram->handle);

	int status=0;
	glGetObjectParameterivARB(mCurProgram->handle,GL_OBJECT_LINK_STATUS_ARB,&status);
	if(!status)
	{
		gLog.OutPut("Failed to Link Program!\n");
#ifdef _ERROR_REPORT
		int infologLength = 0;
		int charsWritten  = 0;
		char *infoLog;
		
		glGetObjectParameterivARB(mCurProgram->handle, GL_OBJECT_INFO_LOG_LENGTH_ARB,
			&infologLength);

		if (infologLength > 0)
		{
			infoLog = (char *)malloc(infologLength);
			glGetInfoLogARB(mCurProgram->handle, infologLength, &charsWritten, infoLog);
			gLog.OutPut(infoLog);
			gLog.OutPut("\n");
			free(infoLog);
		}
#endif
		return false;
	}
	glUseProgramObjectARB(mCurProgram->handle);
	return true;
}
void ShaderMgr::BindProgram(unsigned int idx)
{
	if(idx==0)
	{
		glUseProgramObjectARB(0);
		UTIL_GL::GetGLError();
		return;
	}
	Program* p=NULL;
	if(p=FindProgram(idx))
	{
		glUseProgramObjectARB(p->handle);
		mCurProgram = p;
	}
	else
	{
		glUseProgramObjectARB(0);
	}
	UTIL_GL::GetGLError();

}
void ShaderMgr::DeleteShader(unsigned int shader)
{
	Shader* s=NULL;
	if(s=FindShader(shader))
	{
		glDetachObjectARB(mCurProgram->handle,s->handle);
		glDeleteObjectARB(s->handle);
		vector<Shader>::iterator s;
		for(s=mShaders.begin();s<mShaders.end();s++)
		{
			if((*s).index == shader)
			{
				mShaders.erase(s);
				return;
			}
		}
	}
}
void ShaderMgr::PassUniform1f(char* var, float value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform1fARB(l,value);
}
void ShaderMgr::SetGeometryPrimitive(unsigned int shader, int intype, int outtype, int count)
{
	glProgramParameteriEXT(shader, GL_GEOMETRY_INPUT_TYPE_EXT, intype);
	UTIL_GL::GetGLError();
	glProgramParameteriEXT(shader, GL_GEOMETRY_OUTPUT_TYPE_EXT, outtype);
	UTIL_GL::GetGLError();
	glProgramParameteriEXT(shader, GL_GEOMETRY_VERTICES_OUT_EXT, count);
	UTIL_GL::GetGLError();
}
void ShaderMgr::PassUniform2f(char* var, float value1, float value2)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform2fARB(l,value1,value2);
}
void ShaderMgr::PassUniform3f(char* var, float value1, float value2, float value3)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform3fARB(l,value1,value2,value3);
}
void ShaderMgr::PassUniform4f(char* var, float value1, float value2, float value3, float value4)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform4fARB(l,value1,value2,value3,value4);
}
void ShaderMgr::PassUniform1i(char* var, int value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform1iARB(l,value);
}
void ShaderMgr::PassUniform2i(char* var, int value1, int value2)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform2iARB(l,value1,value2);
}
void ShaderMgr::PassUniform3i(char* var, int value1, int value2, int value3)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform3iARB(l,value1,value2,value3);
}
void ShaderMgr::PassUniform4i(char* var, int value1, int value2, int value3, int value4)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform4iARB(l,value1,value2,value3,value4);
}

void ShaderMgr::PassUniform1fv(char* var, int size, float* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform1fvARB(l,size,value);
}
void ShaderMgr::PassUniform2fv(char* var, int size, float* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform2fvARB(l,size,value);
}
void ShaderMgr::PassUniform3fv(char* var, int size, float* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform3fvARB(l,size,value);
}
void ShaderMgr::PassUniform4fv(char* var, int size, float* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform4fvARB(l,size,value);
}
void ShaderMgr::PassUniform1iv(char* var, int size, int* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform1ivARB(l,size,value);
}
void ShaderMgr::PassUniform2iv(char* var, int size, int* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform2ivARB(l,size,value);
}
void ShaderMgr::PassUniform3iv(char* var, int size, int* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform3ivARB(l,size,value);
}
void ShaderMgr::PassUniform4iv(char* var, int size, int* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniform4ivARB(l,size,value);
}

void ShaderMgr::PassUniformMatrix4f(char* var, int count, bool transpose, float* value)
{
	GLint l;
	if(!mCurProgram)
		return;
	l=glGetUniformLocationARB(mCurProgram->handle,var);
	if(l==-1)
		return;
	glUniformMatrix4fv(l,count,transpose,value);
}

void ShaderMgr::PassAttrib1f(int loc, float value)
{
	glVertexAttrib1fARB(loc,value);
}
void ShaderMgr::PassAttrib2f(int loc, float value1, float value2)
{
	glVertexAttrib2fARB(loc,value1,value2);
}
void ShaderMgr::PassAttrib3f(int loc, float value1, float value2, float value3)
{
	glVertexAttrib3fARB(loc,value1,value2,value3);
}
void ShaderMgr::PassAttrib4f(int loc, float value1, float value2, float value3, float value4)
{
	glVertexAttrib4fARB(loc,value1,value2,value3,value4);
}
void ShaderMgr::PassAttrib1fv(int loc,  float* value)
{
	glVertexAttrib1fvARB(loc,value);
}
void ShaderMgr::PassAttrib2fv(int loc,  float* value)
{
	glVertexAttrib2fvARB(loc,value);
}
void ShaderMgr::PassAttrib3fv(int loc,  float* value)
{
	glVertexAttrib3fvARB(loc,value);
}
void ShaderMgr::PassAttrib4fv(int loc,  float* value)
{
	glVertexAttrib4fvARB(loc,value);
}

void ShaderMgr::PassAttrib4iv(int loc,  int* value)
{
	glVertexAttrib4ivARB(loc,value);
}
#define UNIT(x)	GL_TEXTURE##x##__ARB
void ShaderMgr::SetActiveTextureUnit(int unit)//int UNIT(unit))
{
//	glActiveTexture(UNIT(unit));
	switch(unit)
	{
	case 0:	glActiveTexture(GL_TEXTURE0_ARB);	break;
	case 1:	glActiveTexture(GL_TEXTURE1_ARB);	break;
	case 2:	glActiveTexture(GL_TEXTURE2_ARB);	break;
	case 3:	glActiveTexture(GL_TEXTURE3_ARB);	break;
	}
}
void ShaderMgr::PassSampler1D(char* var, int unit, unsigned int texID)
{
	GLint l = GetUniformLocation(var);
	if(l==-1)
		return;
	glUniform1iARB(l,unit);
	SetActiveTextureUnit(unit);
	glBindTexture(GL_TEXTURE_1D,texID);
}
void ShaderMgr::PassSampler2D(char* var, int unit, unsigned int texID)
{
	GLint l = GetUniformLocation(var);
	if(l==-1)
		return;
	glUniform1iARB(l,unit);

	switch(unit)
	{
	case 0:
		UTIL_GL::BindMultiTexture(GL_TEXTURE0_ARB,texID);
			break;
	case 1:
		UTIL_GL::BindMultiTexture(GL_TEXTURE1_ARB,texID);
			break;
	case 2:
		UTIL_GL::BindMultiTexture(GL_TEXTURE2_ARB,texID);
			break;
	case 3:
		UTIL_GL::BindMultiTexture(GL_TEXTURE3_ARB,texID);
			break;
	}
}
void ShaderMgr::PassSampler3D(char* var, int unit, unsigned int texID)
{
	GLint l = GetUniformLocation(var);
	if(l==-1)
		return;
	glUniform1iARB(l,unit);
	SetActiveTextureUnit(unit);
	glBindTexture(GL_TEXTURE_3D,texID);
}
void ShaderMgr::PassSamplerRect(char* var, int unit, unsigned int texID)
{
	GLint l = GetUniformLocation(var);
	if(l==-1)
		return;
	glUniform1iARB(l,unit);
	SetActiveTextureUnit(unit);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB,texID);
}
void ShaderMgr::PassSamplerCube(char* var, int unit, unsigned int texID)
{
	GLint l = GetUniformLocation(var);
	if(l==-1)
		return;
	glUniform1iARB(l,unit);
	SetActiveTextureUnit(unit);
	glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,texID);
}
GLint ShaderMgr::GetUniformLocation(char* var)
{
	if(!mCurProgram)
		return -1;
	return glGetUniformLocationARB(mCurProgram->handle,var);
}
GLint ShaderMgr::GetAttribLocation(char* var)
{
	if(!mCurProgram)
		return -1;
	return glGetAttribLocationARB(mCurProgram->handle,var);
}
